/* Name: main.c
 * Project: Romboy, a Gameboy ROM dumper
 * Author: Jordan Riggs
 * Creation Date: 2015-07-26
 * Tabsize: 4
 * Copyright: (c) 2015 by Jordan Riggs <jiggs@jiggs.tk>
 * License: GPL
 * This Revision: $Id: main.c,v 1.2 2007/01/14 12:12:27 riggs Exp $
 */




#include "main.h"

//#define DEBUG



//#define RX_BUFFER_SIZE 128
//volatile u08 rx_buf[RX_BUFFER_SIZE];
//volatile u08 rx_buf_pos = 0;
volatile u08 data_avail = 0; // flag to be set by interrupt routine
volatile u08 bytes_received = 0;

u08 waiting_for_ack;
volatile u08 response_code;


volatile u08 rx_packet[PACKETSIZE]; // these could probably be the same array,
u08 tx_packet[PACKETSIZE];          // but it's just easier to keep them separate

cart_header_t header;


void flush()
{
	//rx_buf_pos = 0;
	bytes_received = 0;
	//for (u08 i = 0; i < RX_BUFFER_SIZE; i++) rx_buf[i] = 0;
	//memset(rx_buf, '\0', sizeof(rx_buf));
	for (u08 i = 0; i < PACKETSIZE; i++) rx_packet[i] = 0;
	//memset(rx_packet, '\0', sizeof(rx_packet));
	#ifdef DEBUG
	puts("flushed");
	#endif
}

static int uart_putchar(char c, FILE *stream)
{
	if (c == '\n') uart_putchar('\r', stream);
	loop_until_bit_is_set(UCSRA, UDRE);
	UDR = c;
	return 0;
}

void tx(u08 c)
{
    // easy peasy lemon squeezy!
    while ((UCSRA & _BV(UDRE))==0);
    UDR = c;
}

static FILE debugstdout = FDEV_SETUP_STREAM(uart_putchar,NULL,_FDEV_SETUP_WRITE);

void io_init()
{
	L_ADDR_DDR = 0xFF; // lower address bus
	U_ADDR_DDR = 0xFF; // upper address bus
	DATA_PORT = 0;
	CTRL_DDR |= (BV(GB_RST) | BV(WR) | BV(RD) | BV(CS) | BV(ACT_LED)); // set outputs
	CTRL_DDR &= ~BV(MODE);
	CTRL_PORT |= (BV(GB_RST) | BV(WR) | BV(RD) | BV(CS)); // control lines are active low
	
}

void uart_init()
{
	DDRD &= (1<<1);
	UBRRL = (u08)UART_REG;
	UBRRH = (u08)(UART_REG >> 8);
	//UCSRA = _BV(U2X); //double speed
	UCSRB = (1<<TXEN) | (1<<RXEN) | (1<<RXCIE);     // enable tx+(rx interrupt)
	UCSRC = ((1<<URSEL) | (1<<UCSZ1) | (1<<UCSZ0));  // 8n1
	stdout = &debugstdout;
}

u16 generate_crc16 (u08 packet[PACKETSIZE])
{
  int i;
  u16 crc16 = 0x0000;
  /* generate for whole packet, except for two last bytes */
  for (i = 0; i < PACKETSIZE - 2; i++)
    crc16 = (crc16 << 8) ^ crc16_tab[((crc16 >> 8) ^ packet[i]) & 0xFF];
  return crc16;
}

u08 rd_byte(u16 addr, u08 op)
{
	L_ADDR_PORT = (u08)addr;
	U_ADDR_PORT = (u08)(addr >> 8);
	DATA_DDR = 0x00;
	if (op == RAM) cbi(CTRL_PORT, CS);
	cbi(CTRL_PORT, RD);
	asm volatile("nop");
	asm volatile("nop");
	asm volatile("nop");
	u08 data = DATA_PIN;
	sbi(CTRL_PORT, RD);
	if (op == RAM) sbi(CTRL_PORT, CS);
	return data;
}

void wr_byte(u16 addr, u08 data, u08 op)
{
	L_ADDR_PORT = (u08)addr;
	U_ADDR_PORT = (u08)(addr >> 8);
	DATA_DDR = 0xFF;
	if (op == RAM) cbi(CTRL_PORT, CS);
	cbi(CTRL_PORT, WR);
	asm volatile("nop");
	asm volatile("nop");
	asm volatile("nop");
	DATA_PORT = data;
	sbi(CTRL_PORT, WR);
	if (op == RAM) sbi(CTRL_PORT, CS);
}

u08 verify_logo()
{
	// check cart's nintendo logo against
	// the official nintendo logo
	for (u08 i = 0; i < sizeof(logo); i++)
		if(rd_byte(0x0104+i, ROM) != logo[i]) return 0;
	return 1;
}

void set_mbc1_mode(u08 mode)
{
	u16 addr = 0x6000; // can be anything from 0x6000 to 0x7FFF. MBC1 doesn't care.
	L_ADDR_PORT = (u08)addr;
	U_ADDR_PORT = (u08)(addr >> 8);
	DATA_DDR = 0xFF; // output
	cbi(CTRL_PORT, WR);
	DATA_PORT = mode;
	asm volatile("nop");
	asm volatile("nop");
	asm volatile("nop");
	sbi(CTRL_PORT, WR);
}

void select_rom_bank(u16 bank) // TODO implement other MBCs
{
	if (header.mbc_num == MBC1 && !(bank%0x20)) return; // bank 0 is at 0x0000-0x3FFF; 20h, 40h, 60h not supported
	if (header.mbc_num == MBC3 && !bank) return;  // bank 0 is at 0x0000-0x3FFF
	
	u16 addr = 0x2000;
	L_ADDR_PORT = (u08)addr;
	U_ADDR_PORT = (u08)(addr >> 8);
	DATA_DDR = 0xFF; // output
	DATA_PORT = (u08)(bank&0x00FF);
	cbi(CTRL_PORT, WR);
	asm volatile("nop");
	asm volatile("nop");
	asm volatile("nop");
	sbi(CTRL_PORT, WR);
	
	if (header.mbc_num == MBC1 && bank > 0x1F) // aka bank 31
	{
		set_mbc1_mode(ROM_BANKING); // should be set by default, but this is just to be safe
		addr = 0x4000; // can be anything from 0x4000 to 0x5FFF. MBC1 doesn't care.
		L_ADDR_PORT = (u08)addr;
		U_ADDR_PORT = (u08)(addr >> 8);
		DATA_PORT = (u08)((bank >> 5) & 0x0003); // upper two bits of bank number
		cbi(CTRL_PORT, WR);
		asm volatile("nop");
		asm volatile("nop");
		asm volatile("nop");
		sbi(CTRL_PORT, WR);
	}
	else if (header.mbc_num == MBC5 && bank > 0x00FF) // aka bank 255
	{
		u16 addr = 0x3000; // can be anything from 0x3000 to 0x3FFF
		L_ADDR_PORT = (u08)addr;
		U_ADDR_PORT = (u08)(addr >> 8); // upper bit of bank number
		DATA_PORT = 1;
		cbi(CTRL_PORT, WR);
		asm volatile("nop");
		asm volatile("nop");
		asm volatile("nop");
		sbi(CTRL_PORT, WR);
	}
}

void select_ram_bank(u08 bank) // TODO implement other MBCs
{
	if (header.mbc_num == MBC1)
	{
		if (bank) set_mbc1_mode(RAM_BANKING);
		else return; // default mode is rom banking, with max 8k ram (single bank)
	}
	u16 addr = 0x4000; // can be anything from 0x4000 to 0x5FFF. MBC doesn't care.
	L_ADDR_PORT = (u08)addr;
	U_ADDR_PORT = (u08)(addr >> 8);
	DATA_DDR = 0xFF; // output
	DATA_PORT = bank;
	cbi(CTRL_PORT, WR);
	asm volatile("nop");
	asm volatile("nop");
	asm volatile("nop");
	sbi(CTRL_PORT, WR);
}

// NOTE: This will also enable the RTC registers on MBC3.
void set_ram_state(u08 state)
{
	u16 addr = 0x0000; // can be anything from 0x0000 to 0x1FFF. MBC doesn't care.
	L_ADDR_PORT = (u08)addr;
	U_ADDR_PORT = (u08)(addr >> 8);
	DATA_DDR = 0xFF; // output
	if (state) DATA_PORT = 0x0A; // enable
	else (DATA_PORT) = 0x00;     // disable
	cbi(CTRL_PORT, WR);
	asm volatile("nop");
	asm volatile("nop");
	asm volatile("nop");
	sbi(CTRL_PORT, WR);
}

void rd_header()
{
	#ifdef DEBUG
	puts("reading header");
	#endif
	for (u08 i = 0; i < 16; i++)
		header.title[i] = rd_byte(i + 0x0134, ROM);
	header.cgb = header.title[15] & 0x80;
	header.sgb = rd_byte(0x0146, ROM);
	header.mbc_type = rd_byte(0x0147, ROM);
	header.mbc_num = identify_mbc(header.mbc_type);
	header.rom_size = rd_byte(0x0148, ROM);
	if (header.rom_size & 0xF0) // 1.1, 1.2, and 1.5 MB ROMs are coded as 0x52, 0x53, and 0x54
		header.num_rom_banks = (2 << ((header.rom_size & 0xF0)/16)) + (2 << (header.rom_size & 0x0F));
	else						// all others are coded as 0x00 (2 banks) to 0x07 (256 banks)
		header.num_rom_banks = 2 << (header.rom_size & 0x0F);
	header.ram_size = rd_byte(0x0149, ROM);
	if (header.ram_size == 1 || header.ram_size == 2)
		header.num_ram_banks = 1;
	else if (header.ram_size == 3)
		header.num_ram_banks = 4;
	else if (header.ram_size == 4)
		header.num_ram_banks = 16;
	else header.num_ram_banks = 0;
	header.header_checksum = rd_byte(0x014D, ROM);
	// verify header checksum
	short check = 0;
	for(u16 addr = 0x0134; addr <= 0x014C; addr++)
		check = check - rd_byte(addr, ROM) - 1;
	header.header_checksum_valid = ((check & 0xFF) == header.header_checksum);
	header.global_checksum = 256 * rd_byte(0x014E, ROM) + rd_byte(0x014F, ROM);
	header.header_received = 1;
}

mbc_t identify_mbc(u08 type)
{
	switch (type)
	{
	case 0x00:
		return MBC_NONE;
		break;
	case 0x01:
		return MBC1;
		break;
	case 0x02:
		return MBC1;
		break;
	case 0x03:
		return MBC1;
		break;
	case 0x05:
		return MBC2;
		break;
	case 0x06:
		return MBC2;
		break;
	case 0x08:
		return MBC_UNKNOWN;
		break;
	case 0x09:
		return MBC_UNKNOWN;
		break;
	case 0x0B:
		return MBC_UNKNOWN;
		break;
	case 0x0C:
		return MBC_UNKNOWN;
		break;
	case 0x0D:
		return MBC_UNKNOWN;
		break;
	case 0x0F:
		return MBC3;
		break;
	case 0x10:
		return MBC3;
		break;
	case 0x11:
		return MBC3;
		break;
	case 0x12:
		return MBC3;
		break;
	case 0x13:
		return MBC3;
		break;
	case 0x15:
		return MBC4;
		break;
	case 0x16:
		return MBC4;
		break;
	case 0x17:
		return MBC4;
		break;
	case 0x19:
		return MBC5;
		break;
	case 0x1A:
		return MBC5;
		break;
	case 0x1B:
		return MBC5;
		break;
	case 0x1C:
		return MBC5;
		break;
	case 0x1D:
		return MBC5;
		break;
	case 0x1E:
		return MBC5;
		break;
	case 0xFC:
		return MBC_UNKNOWN;
		break;
	case 0xFD:
		return MBC_UNKNOWN;
		break;
	case 0xFE:
		return MBC_UNKNOWN;
		break;
	case 0xFF:
		return MBC_UNKNOWN;
		break;
	}
	
	return MBC_UNKNOWN;
}

void send_header_info()
{
	tx(DATA); // data
	tx(0x00); // ?
	tx(VERSION_MAJOR); // hw version
	tx(VERSION_MINOR); // hw version
	tx(0x00); // mfr id
	tx(0x00); // chip id
	tx(0x00); // bbl flag
	tx(0x00); // ?
	tx(verify_logo()); // logo_correct
	for (u08 i = 0; i < sizeof(header.title); i++)
		tx(header.title[i]); // title
	tx(0x00); // ?
	tx(0x00); // ?
	tx(header.sgb); // sgb flag
	tx(header.mbc_type); // mbc type
	tx(header.rom_size); // rom size
	tx(header.ram_size); // ram size
	tx(0x00); // ?
	tx(0x00); // ?
	tx(header.header_checksum_valid); // ?
	tx(header.header_checksum); // ?
	tx((u08)((header.global_checksum & 0xFF00) >> 8)); // global checksum (upper)
	tx((u08)(header.global_checksum & 0x00FF)); // global checksum (lower)
	for (u08 i = 37; i < PACKETSIZE; i++) tx(0x00); // fill the rest of the packet
}

void send_rom_packet(u16 bank, u08 frame) // 256 frames per 16k bank
{
	tx_packet[0] = DATA;
	if (bank == header.num_rom_banks-1 && frame == 255)
		tx_packet[1] = LAST_DATA;
	else
		tx_packet[1] = NORMAL_DATA;
	tx_packet[2] = 0x00;
	tx_packet[3] = frame;
	tx_packet[4] = bank&0x0100;
	tx_packet[5] = (u08)(bank&0xFF);
	
	if (!bank)
	{
		for (u08 i = 0; i < FRAMESIZE; i++) tx_packet[6+i] = rd_byte(i+(FRAMESIZE*frame), ROM);
	}
	else
	{
		select_rom_bank(bank);
		for (u08 i = 0; i < FRAMESIZE; i++) tx_packet[6+i] = rd_byte(0x4000+(FRAMESIZE*frame)+i, ROM);
	}
	
	u16 crc = generate_crc16(tx_packet);
	
	tx_packet[PACKETSIZE-2] = (u08)((crc & 0xFF00) >> 8);
	tx_packet[PACKETSIZE-1] = (u08)(crc & 0x00FF);
	
	for (u08 i = 0; i < PACKETSIZE; i++) tx(tx_packet[i]);
}

void send_ram_packet(u08 bank, u08 frame) // 128 frames per 8k bank
{
	tx_packet[0] = DATA;
	if (header.rom_size == 1 && frame == 31) // only have 2k to send
		tx_packet[1] = LAST_DATA;
	else if (bank == header.num_ram_banks-1 && frame == 127)
		tx_packet[1] = LAST_DATA;
	else
		tx_packet[1] = NORMAL_DATA;
	tx_packet[2] = 0x00;
	tx_packet[3] = frame;
	tx_packet[4] = 0x00;
	tx_packet[5] = bank;
	select_ram_bank(bank);	
	for (u08 i = 0; i < FRAMESIZE; i++) tx_packet[6+i] = rd_byte(0xA000+(FRAMESIZE*frame)+i, RAM);
	
	u16 crc = generate_crc16(tx_packet);
	tx_packet[PACKETSIZE-2] = (u08)((crc & 0xFF00) >> 8);
	tx_packet[PACKETSIZE-1] = (u08)(crc & 0x00FF);
	
	for (u08 i = 0; i < PACKETSIZE; i++) tx(tx_packet[i]);
}

void fill_ram(u08 byte)
{
	ACT_LED_ON;
	set_ram_state(ENABLE);
	u08 num_frames = 8192 / FRAMESIZE; // frames per page of ram
	if (header.rom_size == 1) num_frames /= 4; // only 2k ram
	for(u08 bank = 0; bank < header.num_ram_banks; bank++)
	{
		select_ram_bank(bank);
		for(u08 frame = 0; frame < num_frames; frame++)
			for(u08 i = 0; i < FRAMESIZE; i++)
				for (u08 i = 0; i < FRAMESIZE; i++) wr_byte(0xA000+i+(FRAMESIZE*frame), byte, RAM);
	}
	set_ram_state(DISABLE);
	ACT_LED_OFF;
}

int main(void)
{
	uart_init();
	io_init();
	sei();
    //DEBUGF("two plus to is: %d\n", 4);
    u08 control_char, packet_type, op_type;
	while (1)
	{
		if (data_avail)
		{
			data_avail = 0;
			//rx_packet[bytes_received-1] = rx_buf[rx_buf_pos - 1];
			//if (bytes_received == PACKETSIZE) // do we have a complete packet?
			//{
				#ifdef DEBUG
				puts("got a packet");
				for (u08 i = 0; i < PACKETSIZE; i++) tx(rx_packet[i]); puts("");
				#endif
				control_char = rx_packet[0];
				packet_type =  rx_packet[1];
				op_type =      rx_packet[2];
				if (control_char == DATA)
				{
					#ifdef DEBUG
					puts("DATA");
					#endif
					if (packet_type == STATUS)
					{
						#ifdef DEBUG
						puts("STATUS");
						#endif
						ACT_LED_ON;
						rd_header();
						ACT_LED_OFF;
						send_header_info();
						#ifdef DEBUG
						puts("\nheader sent");
						#endif
					}
					else if (packet_type == CONFIG)
					{
						#ifdef DEBUG
						puts("CONFIG");
						#endif
						if (op_type == RROM)
						{
							#ifdef DEBUG
							puts("RROM");
							#endif
							ACT_LED_ON;
							rd_header(); // in case this hasn't already been done
							for (u16 bank = 0; bank < header.num_rom_banks; bank++)
							{
							
								if (header.mbc_num == MBC1 && bank > 0 && !(bank%0x20))
								{
									#ifdef DEBUG
									puts("skipping a bank");
									#endif
									continue; // MBC1 isn't campatible with banks 20h, 40h, 60h
								}
								for (u16 frame = 0; frame <= 255; frame++)
								{
									send_rom_packet(bank, (u08)frame);
									waiting_for_ack = 1;
									while (response_code != ACK) asm("nop"); // TODO set a watchdog timer, check for NAK
									response_code = 0x00;
									flush();
									waiting_for_ack = 0;
								}
							}
							ACT_LED_OFF;
							#ifdef DEBUG
							puts("rom packet sent");
							#endif
						}
						else if (op_type == RRAM)
						{
							#ifdef DEBUG
							puts("RRAM");
							#endif
							ACT_LED_ON;
							rd_header(); // in case this hasn't already been done
							if (header.ram_size == 0 && header.mbc_num != MBC2) // MBC2 has built-in RAM
							{
								tx(NAK);
								continue; // no ram available, so ignore request
							}
							else
							{
								set_ram_state(ENABLE);
								u08 num_frames;
								if (header.ram_size == 1) num_frames = 32; // accomodate 2k ram
								else num_frames = 128; // all other sizes have 8k banks
								for (u08 bank = 0; bank < header.num_ram_banks; bank++)
								{
									for (u08 frame = 0; frame < num_frames; frame++)
									{
										send_ram_packet(bank, frame);
										waiting_for_ack = 1;
										while (response_code != ACK) asm("nop"); // TODO set a watchdog timer, check for NAK
										response_code = 0x00;
										flush();
										waiting_for_ack = 0;
									}
								}
								set_ram_state(DISABLE);
							}
							ACT_LED_OFF;
						}
						else if (op_type == WROM)
						{
							#ifdef DEBUG
							puts("WROM");
							#endif
							ACT_LED_ON;
							ACT_LED_OFF;
						}
						else if (op_type == WRAM)
						{
							#ifdef DEBUG
							puts("WRAM");
							#endif
							rd_header(); // in case this hasn't already been done
							//ACT_LED_OFF;
							if (header.rom_size == 0 && header.mbc_num != MBC2)
							{
								tx(NAK);
								continue; // no ram available, so ignore request
							}
							else
							{
								flush();
								tx(ACK);
								ACT_LED_ON;
								set_ram_state(ENABLE);
								u08 num_frames = 8192 / FRAMESIZE; // frames per page of ram
								if (header.rom_size == 1) num_frames /= 4; // only 2k ram
								for(u08 bank = 0; bank < header.num_ram_banks; bank++)
								{
									select_ram_bank(bank);
									for(u08 frame = 0; frame < num_frames; frame++)
									{
										// wait for the next packet to arrive
										while (!data_avail) asm("nop"); // TODO set a watchdog timer, check for correct frame number
										data_avail = 0;
										for(u08 i = 0; i < FRAMESIZE; i++)
										{
											for (u08 i = 0; i < FRAMESIZE; i++) wr_byte(0xA000+i+(FRAMESIZE*frame), rx_packet[6+i], RAM);
										}
										flush();
										tx(ACK);
									}
								}
								set_ram_state(DISABLE);
								ACT_LED_OFF;
							}
						}
					}
					else if (packet_type == ERASE)
					{
						if (op_type == EFLA)
						{
							#ifdef DEBUG
							puts("EFLA");
							#endif
						}
						else if (op_type == ERAM)
						{
							#ifdef DEBUG
							puts("ERAM");
							#endif
							fill_ram(0x00);
							tx(ACK);
						}
					}
				}
				// else if (control_char == WHATEVER)
				
				flush(); // clear the rx buffer after we've done something with the received packet
			//}
			// else chill
		}
	}
	return 0;
}

ISR(USART_RXC_vect)
{
	u08 rx_byte = UDR;
	
	if (waiting_for_ack)
	{
		response_code = rx_byte;
		return;
	}
	
	if (bytes_received < PACKETSIZE)
	{
		rx_packet[bytes_received] = rx_byte;
		if (++bytes_received == PACKETSIZE) data_avail = 1;
		// any further packets received are ignored until flush() is called
	}
	
	//rx_buf[rx_buf_pos] = rx_byte;
	//rx_buf_pos = ++rx_buf_pos%RX_BUFFER_SIZE;
}

